/*
 * <<< mp_bitonic.c >>>
 *
 * --- Sample application for isis 'bitonic sort' - for MPI
 *     Copyright (C) 2000-2001 Amano Lab., Keio University. ---
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <mpi.h>

#define VALUEBIT		16
#define VALUEMASK		(~(~0 << VALUEBIT))
#define RADIXBIT		(VALUEBIT >> 1)
#define DEFAULT_SIZE	64
#define MAX_PRINT_SIZE	256

typedef unsigned int data_t;

#define rnd() ((unsigned)((seed = 1566083941UL * seed + 1) >> 16))
#define srnd(x) (seed = (x))
static unsigned long seed = 1;

static void show_help(void);
static int fix_number_to_exp2n(int);
static void init(data_t*, int);
static void show(data_t*, int);
static int check(data_t*, int);
static void radix_sort(data_t*, data_t*, int);
static void radix_rev_sort(data_t*, data_t*, int);
static void bitonic_split(data_t*, data_t*, int);
static void bitonic_rev_split(data_t*, data_t*, int);
static void bitonic_to_sequence(data_t*, data_t*, int);
static void bitonic_to_rev_sequence(data_t*, data_t*, int);
static void bitonic_split_with_comm(data_t*, data_t*, int, int, int);
static void bitonic_rev_split_with_comm(data_t*, data_t*, int, int, int);
void bitonic_sort(data_t*, data_t*, int, int, int, int, int);

void show_help(void)
{
	puts("usage: mp_bitonic [options] [size]");
	puts("options:");
	puts("  -r<n>  set threshold of radix sort.");
	puts("  -h     print this message.");
	puts("  -i     print timing statistics.");
	puts("  -t     check to make sure all keys are sorted correctly.");
	puts("  -v     verbosely output.");
}

int fix_number_to_exp2n(int x)
{
	int i = 0;
	while (x != 0) {
		x >>= 1;
		i++;
	}
	return (i == 0) ? 0 : (1 << (i - 1));
}

void init(data_t *a, int size)
{
	int	i;
	for (i = 0; i < size; i++) {
		unsigned long rnd_num = ((unsigned long)rnd() << 16) | rnd();
		a[i] = ((rnd_num >> (32 - VALUEBIT)) & VALUEMASK);
	}
}

void show(data_t *a, int size)
{
	int i;
	for (i = 0; ; i++) {
		printf("%0*x", ((VALUEBIT + 3) / 4), a[i]);
		if (i == size - 1) break;
		putchar('.');
	}
}

int check(data_t *a, int size)
{
	int i;
	for (i = 1; i < size; i++) {
		if (a[i] < a[i - 1]) return 0;
	}
	return 1;
}

void radix_sort(data_t *a, data_t *w, int n)
{
	static int c[1 << RADIXBIT];
	const int radix = (1 << RADIXBIT), radixmask = radix - 1;
	int i, j;
#ifdef DEBUG
	printf("%02d/%02d radix_sort(n:%d):\n", get_puid(), get_punum(), n);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf("\n");
#endif /* DEBUG */
	for (i = 0; i < radix; i++) c[i] = 0;
	for (i = 0; i < n; i++) c[a[i] & radixmask]++;
	for (j = c[0], i = 1; i < radix; i++) j = (c[i] += j);
	for (i = n - 1; i >= 0; i--) w[--c[a[i] & radixmask]] = a[i];
	for (i = 0; i < radix; i++) c[i] = 0;
	for (i = 0; i < n; i++) c[w[i] >> RADIXBIT]++;
	for (j = c[0], i = 1; i < radix; i++) j = (c[i] += j);
	for (i = n - 1; i >= 0; i--) a[--c[w[i] >> RADIXBIT]] = w[i];
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf("\n");
#endif /* DEBUG */
}

void radix_rev_sort(data_t *a, data_t *w, int n)
{
	static int c[1 << RADIXBIT];
	const int radix = (1 << RADIXBIT), radixmask = radix - 1;
	int i, j;
#ifdef DEBUG
	printf("%02d/%02d radix_rev_sort(n:%d):\n", get_puid(), get_punum(), n);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf("\n");
#endif /* DEBUG */
	for (i = 0; i < radix; i++) c[i] = 0;
	for (i = 0; i < n; i++) c[a[i] & radixmask]++;
	for (j = c[radix - 1], i = radix - 2; i >= 0; i--) j = (c[i] += j);
	for (i = n - 1; i >= 0; i--) w[--c[a[i] & radixmask]] = a[i];
	for (i = 0; i < radix; i++) c[i] = 0;
	for (i = 0; i < n; i++) c[w[i] >> RADIXBIT]++;
	for (j = c[radix - 1], i = radix - 2; i >= 0; i--) j = (c[i] += j);
	for (i = n - 1; i >= 0; i--) a[--c[w[i] >> RADIXBIT]] = w[i];
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf("\n");
#endif /* DEBUG */
}

void bitonic_split(data_t *a, data_t *b, int n)
{
	int i;
#ifdef DEBUG
	printf("%02d/%02d bitonic_split(n:%d):\n", get_puid(), get_punum(), n);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
	for (i = 0; i < n; i++) {
		if (a[i] > b[i]) {
			data_t tmp;
			tmp = a[i]; a[i] = b[i]; b[i] = tmp;
		}
	}
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
}

void bitonic_rev_split(data_t *a, data_t *b, int n)
{
	int i;
#ifdef DEBUG
	printf("%02d/%02d bitonic_rev_split(n:%d):\n", get_puid(), get_punum(), n);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
	for (i = 0; i < n; i++) {
		if (a[i] < b[i]) {
			data_t tmp;
			tmp = a[i]; a[i] = b[i]; b[i] = tmp;
		}
	}
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
}

void bitonic_to_sequence(data_t *a, data_t *w, int n)
{
	int i, j, k, inc_flag;
#ifdef DEBUG
	printf("%02d/%02d bitonic_to_sequence(n:%d):\n", get_puid(), get_punum(),
		   n);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf("\n");
#endif /* DEBUG */
	/* search maximum and minimum point */
	for (i = 0; i < n - 2; i++) {
		if (a[i] < a[i + 1]) {
			inc_flag = 1;
			break;
		} else if (a[i] > a[i + 1]) {
			inc_flag = 0;
			break;
		}
	}
	if (i == n - 2) return;
	if (inc_flag) {
		/* first direction is increase */
		while (a[i] <= a[i + 1]) {
			i++;
			if (i == n - 1) return;
		}
		j = 0; while (i < n) { w[j] = a[i]; i++, j++; }
		i = 0; while (j < n) { w[j] = a[i]; i++, j++; }
		i = 0, j = n - 1;
		for (k = n - 1; k >= 0; k--) {
			if (w[i] >= w[j]) {
				a[k] = w[i]; i++;
			} else {
				a[k] = w[j]; j--;
			}
		}
	} else {
		/* first direction is decrease */
		while (a[i] >= a[i + 1]) {
			i++;
			if (i == n - 1) break;
		}
		j = 0; while (i < n) { w[j] = a[i]; i++, j++; }
		i = 0; while (j < n) { w[j] = a[i]; i++, j++; }
		i = 0, j = n - 1;
		for (k = 0; k < n; k++) {
			if (w[i] <= w[j]) {
				a[k] = w[i]; i++;
			} else {
				a[k] = w[j]; j--;
			}
		}
	}
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf("\n");
#endif /* DEBUG */
}

void bitonic_to_rev_sequence(data_t *a, data_t *w, int n)
{
	int i, j, k, inc_flag;
#ifdef DEBUG
	printf("%02d/%02d bitonic_to_rev_sequence(n:%d):\n",
		   get_puid(), get_punum(), n);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf("\n");
#endif /* DEBUG */
	/* search maximum and minimum point */
	for (i = 0; i < n - 2; i++) {
		if (a[i] < a[i + 1]) {
			inc_flag = 1;
			break;
		} else if (a[i] > a[i + 1]) {
			inc_flag = 0;
			break;
		}
	}
	if (i == n - 2) return;
	if (inc_flag) {
		/* first direction is increase */
		while (a[i] <= a[i + 1]) {
			i++;
			if (i == n - 1) break;
		}
		j = 0; while (i < n) { w[j] = a[i]; i++, j++; }
		i = 0; while (j < n) { w[j] = a[i]; i++, j++; }
		i = 0, j = n - 1;
		for (k = 0; k < n; k++) {
			if (w[i] >= w[j]) {
				a[k] = w[i]; i++;
			} else {
				a[k] = w[j]; j--;
			}
		}
	} else {
		/* first direction is decrease */
		while (a[i] >= a[i + 1]) {
			i++;
			if (i == n - 1) return;
		}
		j = 0; while (i < n) { w[j] = a[i]; i++, j++; }
		i = 0; while (j < n) { w[j] = a[i]; i++, j++; }
		i = 0, j = n - 1;
		for (k = n - 1; k >= 0; k--) {
			if (w[i] <= w[j]) {
				a[k] = w[i]; i++;
			} else {
				a[k] = w[j]; j--;
			}
		}
	}
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf("\n");
#endif /* DEBUG */
}

int main(int argc, char **argv)
{
	data_t *a;
	data_t *w;
	clock_t init_start_time, init_end_time, calc_start_time, calc_end_time;
	size_t size = DEFAULT_SIZE, radix_threshold = 0, punum, puid, n;
	int check_flag = 0, info_flag = 0, verbose_flag = 0, result;

	/* initialize MPI environment and random seed */
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &punum);
	MPI_Comm_rank(MPI_COMM_WORLD, &puid);
	punum = fix_number_to_exp2n(punum);
	if (puid >= punum) return 0;
	srnd(puid + 1); rnd();

	/* read arguments */
	while (*++argv != NULL) {
		if (**argv == '-') {
			switch (*++*argv) {
			case 'h':
				if (puid == 0) show_help();
				return 0;
			case 'i':
				info_flag = 1;
				break;
			case 'r':
				radix_threshold = atoi(*argv + 1);
				break;
			case 't':
				check_flag = 1;
				break;
			case 'v':
				verbose_flag = 1;
				break;
			default:
				break;
			}
		} else if (isdigit((int)**argv)) {
			size = atoi(*argv);
			if (size < 0) size = DEFAULT_SIZE;
		}
	}
	size = fix_number_to_exp2n(size);
	if (size < 4) size = 4;
	if (punum > size / 4) {
		if (puid == 0) fprintf(stderr, "Too many processors.\n");
		exit(1);
	}
	n = size / punum;
	radix_threshold = fix_number_to_exp2n(radix_threshold);
	if (radix_threshold == 0 || radix_threshold > n) radix_threshold = n;
	if (radix_threshold > size / 2) radix_threshold = size / 2;
	if (radix_threshold < 4) radix_threshold = 4;
	if (verbose_flag && puid == 0) {
		printf("size:%d radix-threshold:%d value:0-%#x\n",
			   size, radix_threshold, VALUEMASK);
	}
	init_start_time = init_end_time = calc_start_time = calc_end_time = 0;

	/* initialize */
	a = (data_t*)malloc(n * sizeof(data_t));
	w = (data_t*)malloc(n * sizeof(data_t));
	if (a == NULL || w == NULL) {
		fputs("Out of memory.\n", stderr);
		exit(1);
	}
	if (info_flag) init_start_time = clock();
#ifndef DEBUG
	init(a, n);
#else
	/* simple init data: for debug only */
	{
		int i;
		for (i = 0; i < n; i++) a[i] = (i + puid * n) & VALUEMASK;
	}
#endif /* DEBUG */
	if (info_flag) init_end_time = clock();
	if (verbose_flag && size <= MAX_PRINT_SIZE) {
		printf("%02d: ", puid);
		show(a, n);
		printf("\n");
	}

	/* calculate */
	if (info_flag) calc_start_time = clock();
	bitonic_sort(a, w, size, n, radix_threshold, punum, puid);
	if (info_flag) calc_end_time = clock();

	/* check */
	if (check_flag) {
		int value[2];
		result = check(a, n);
		if (puid == 0) {
			MPI_Status dummy;
			int max = a[n - 1], i;
			for (i = 1; i < punum; i++) {
				MPI_Recv(value, 2, MPI_INT, i, 0, MPI_COMM_WORLD, &dummy);
				if (max > value[0] || value[0] > value[1]) {
					result = 0;
				}
				max = value[1];
			}
			if (verbose_flag) {
				puts(result ? "success." : "failed.");
			}
		} else {
			if (result) {
				value[0] = a[0];
				value[1] = a[n - 1];
			} else {
				value[0] = 1, value[1] = 0;
			}
			MPI_Send(value, 2, MPI_INT, 0, 0, MPI_COMM_WORLD);
		}
	} else {
		result = 1;
	}

	/* show results */
	if (verbose_flag && size <= MAX_PRINT_SIZE) {
		printf("%02d: ", puid);
		show(a, n);
		printf("\n");
	}

	/* show timing statistics */
	if (info_flag && puid == 0) {
		printf("init start: %10ld\n"
			   "init end:   %10ld\n"
			   "calc start: %10ld\n"
			   "calc end:   %10ld\n"
			   "init time:  %10ld\n"
			   "calc time:  %10ld\n",
			   (long)init_start_time, (long)init_end_time,
			   (long)calc_start_time, (long)calc_end_time,
			   (long)(init_end_time - init_start_time),
			   (long)(calc_end_time - calc_start_time));
	}

	MPI_Finalize();
	return (result ? 0 : 1);
}

void bitonic_split_with_comm(data_t *a, data_t *b, int n, int partner, int tag)
{
	MPI_Status dummy;
#ifdef DEBUG
	printf("%02d/%02d comm with pu%02d:\n", get_puid(), get_punum(), partner);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
	MPI_Send(b, n, MPI_INT, partner, tag, MPI_COMM_WORLD);
	MPI_Recv(b, n, MPI_INT, partner, tag, MPI_COMM_WORLD, &dummy);
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
	bitonic_split(a, b, n);
#ifdef DEBUG
	printf("%02d/%02d comm with pu%02d:\n", get_puid(), get_punum(), partner);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
	MPI_Send(b, n, MPI_INT, partner, tag + 1, MPI_COMM_WORLD);
	MPI_Recv(b, n, MPI_INT, partner, tag + 1, MPI_COMM_WORLD, &dummy);
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
}

void bitonic_rev_split_with_comm(data_t *a, data_t *b, int n, int partner,
								 int tag)
{
	MPI_Status dummy;
#ifdef DEBUG
	printf("%02d/%02d comm with pu%02d:\n", get_puid(), get_punum(), partner);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
	MPI_Send(b, n, MPI_INT, partner, tag, MPI_COMM_WORLD);
	MPI_Recv(b, n, MPI_INT, partner, tag, MPI_COMM_WORLD, &dummy);
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
	bitonic_rev_split(a, b, n);
#ifdef DEBUG
	printf("%02d/%02d comm with pu%02d:\n", get_puid(), get_punum(), partner);
	printf("%02d/%02d < ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
	MPI_Send(b, n, MPI_INT, partner, tag + 1, MPI_COMM_WORLD);
	MPI_Recv(b, n, MPI_INT, partner, tag + 1, MPI_COMM_WORLD, &dummy);
#ifdef DEBUG
	printf("%02d/%02d > ", get_puid(), get_punum());
	show(a, n); printf(" : "); show(b, n); printf("\n");
#endif /* DEBUG */
}

void bitonic_sort(data_t *a, data_t *w, int size, int n, int radix_threshold,
				  int punum, int puid)
{
	/* sort data in each processor: there is no communication */
	if (radix_threshold < n) {
		int i;
		for (i = 0; i < n; /* nothing to do */ ) {
			radix_sort(a + i, w, radix_threshold);
			i += radix_threshold;
			radix_rev_sort(a + i, w, radix_threshold);
			i += radix_threshold;
		}
		for (i = (radix_threshold << 1); i < n; /* nothing to do */ ) {
			const int i_mul_2 = (i << 1);
			int j;
			for (j = i; j > radix_threshold; /* nothing to do */ ) {
				const j_div_2 = (j >> 1);
				int k;
				for (k = 0; k < n; k += i_mul_2) {
					int l;
					for (l = k; l < i + k; l += j) {
						bitonic_split(a + l, a + l + j_div_2, j_div_2);
						bitonic_rev_split(a + i + l, a + i + l + j_div_2,
										  j_div_2);
					}
				}
				j = j_div_2;
			}
			for (j = 0; j < n; j += i_mul_2) {
				int k;
				for (k = j; k < i + j; k += radix_threshold) {
					bitonic_to_sequence(a + k, w, radix_threshold);
					bitonic_to_rev_sequence(a + i + k, w, radix_threshold);
				}
			}
			i = i_mul_2;
		}
		if ((puid & 1) == 0) {
			for (i = n; i > radix_threshold; /* nothing to do */ ) {
				const int i_div_2 = (i >> 1);
				int j;
				for (j = 0; j < n; j += i) {
					bitonic_split(a + j, a + j + i_div_2, i_div_2);
				}
				i = i_div_2;
			}
			for (i = 0; i < n; i += radix_threshold) {
				bitonic_to_sequence(a + i, w, radix_threshold);
			}
		} else {
			for (i = n; i > radix_threshold; /* nothing to do */ ) {
				const int i_div_2 = (i >> 1);
				int j;
				for (j = 0; j < n; j += i) {
					bitonic_rev_split(a + j, a + j + i_div_2, i_div_2);
				}
				i = i_div_2;
			}
			for (i = 0; i < n; i += radix_threshold) {
				bitonic_to_rev_sequence(a + i, w, radix_threshold);
			}
		}
	} else {
		if ((puid & 1) == 0) {
			radix_sort(a, w, n);
		} else {
			radix_rev_sort(a, w, n);
		}
	}
#ifdef DEBUG
	printf("%02d/%02d: ", puid, punum); show(a, n); printf("\n");
#endif /* DEBUG */
	/* sort whole data with inter-processor communications */
	{
		const size_t commsize = (n >> 1);
		int tag = 0;
		int i, punum_shift; /* punum == (1 << punum_shift) */
		punum_shift = 0;
		i = punum;
		while (i > 1) {
			i >>= 1;
			punum_shift++;
		}
		for (i = 0; i < punum_shift; i++) {
			if (((puid >> i) & 2) == 0) {
				int j;
				for (j = i; j >= 0; j--) {
					int partner = (puid ^ (1 << j));
					if (puid < partner) {
						bitonic_split_with_comm(a, a + commsize, commsize,
												partner, tag);
					} else {
						bitonic_rev_split_with_comm(a + commsize, a, commsize,
													partner, tag);
					}
					tag += 2;
				}
				for (j = n; j > radix_threshold; /* nothing to do */ ) {
					const int j_div_2 = (j >> 1);
					int k;
					for (k = 0; k < n; k += j) {
						bitonic_split(a + k, a + k + j_div_2, j_div_2);
					}
					j = j_div_2;
				}
				for (j = 0; j < n; j += radix_threshold) {
					bitonic_to_sequence(a + j, w, radix_threshold);
				}
			} else {
				int j;
				for (j = i; j >= 0; j--) {
					int partner = (puid ^ (1 << j));
					if (puid < partner) {
						bitonic_rev_split_with_comm(a + commsize, a, commsize,
													partner, tag);
					} else {
						bitonic_split_with_comm(a, a + commsize, commsize,
												partner, tag);
					}
					tag += 2;
				}
				for (j = n; j > radix_threshold; /* nothing to do */ ) {
					const int j_div_2 = (j >> 1);
					int k;
					for (k = 0; k < n; k += j) {
						bitonic_rev_split(a + k, a + k + j_div_2, j_div_2);
					}
					j = j_div_2;
				}
				for (j = 0; j < n; j += radix_threshold) {
					bitonic_to_rev_sequence(a + j, w, radix_threshold);
				}
			}
#ifdef DEBUG
			printf("%02d/%02d: ", puid, punum); show(a, n); printf("\n");
#endif /* DEBUG */
		}
	}
}
